//////////////////////////////
// The Hacker Mindset Badge //
//      by Garrett Gee      //
//////////////////////////////
// Build instructions:
// arduino app 1.8.x required (littlefs uploader doesn't work in 2.x series)
// 
// Add board manager (file -> prefs) 
//    https://github.com/earlephilhower/arduino-pico/releases/download/global/package_rp2040_index.json
// Add libraries (tools -> manage libraries)
//    adafruit neopixel
//    ws2812fx
//    SafeString
//    RBD_Capacitance
// Add libraries manually
//    https://github.com/earlephilhower/arduino-pico-littlefs-plugin/releases/
// Compile options (tools menu)
//    Board: Generic RP2040
//    Flash Size "2MB" (1mb sketch, 1mb fs) (larger setting will just take longer)
//    CPU Speed "100Mhz" - crashes less than 50Mhz
//    Boot stage 2: "W25Q64JV QSPI/4"
//    USB stack - pico sdk
// Steps
//    1. sketch upload
//    2. littefs data upload
//    note: should take ~1 minute for both tasks
//  
// ----------------------------------------------------------------------

// libraries setup
#ifdef __AVR__
 #include <avr/power.h> // Required for 16 MHz Adafruit Trinket
#endif

#include <EEPROM.h>
#include <Keyboard.h>
#include <Mouse.h>
#include <LittleFS.h>
#include <SingleFileDrive.h>

#include <Adafruit_NeoPixel.h>
#include <WS2812FX.h>
#include <millisDelay.h>
#include <RBD_Capacitance.h>

#include "blinky.h"
#include "ducky.h"

// button setup
int LEFT_PIN = 20; //Buttons are Active High
int RIGHT_PIN =  7; //Buttons are Active High
int cptchSnd    = 19; //for capacitive touch
int cptchRcv    = 18; //for capacitive touch

// LED setup
#define LED_PIN    5
#define LED_COUNT 14
Adafruit_NeoPixel pixels = Adafruit_NeoPixel(LED_COUNT, LED_PIN, NEO_GRB + NEO_KHZ800);
WS2812FX ws2812fx = WS2812FX(LED_COUNT, LED_PIN, NEO_GRB + NEO_KHZ800);
millisDelay ledDelay;
millisDelay ledrotateDelay;
byte favs[] = { 10, 18, 20, 33, 42, 7, 43, 8 };
byte favs_cnt = 7; //minus 1 bc start at 0
byte favs_current = 0;
byte isLEDmode = 0;

// capacitive touch setup
RBD::Capacitance cap_sensor(cptchSnd, cptchRcv); // send, receive pin
millisDelay cptchbaselinetimer;
int samplcount = 0;
int sampltotal = 0;
int curval;
//int baseval=50; //50-51 seems baseline after stable. but seen 55-62 on boot??
//int baseval;
//int valthres; //how much it needs to change
int curpress=0;
int pressthres=20; //how many instances of it needs to be changed into order to set button change mode. 20 seems to be a decent tap/press from me. quick tap doesn't trigger
bool baselinerun = false;

// variables for button logic
long debouncing_time = 250;
unsigned long last_micros = 0;
volatile byte btncounter = 0;
volatile byte btnid  = 0;
volatile byte l_count = 0;
volatile byte r_count = 0;

// kb setup
byte isKBloop = 0;
byte jigglerstage = 0;
millisDelay mousejigglerDelay;

// serial setup
#define MAX_NUM_CHARS 16 // maximum number of characters read from the Serial Monitor
char scmd[MAX_NUM_CHARS];    // char[] to store incoming serial commands
bool scmd_complete = false;  // whether the command string is complete

// usb drive setup
uint32_t cnt = 0;
bool okayToWrite = true;
// Called when the USB stick connected to a PC and the drive opened
// Note this is from a USB IRQ so no printing to SerialUSB/etc.
void plug(uint32_t i) {
  (void) i;
  okayToWrite = false;
}
// Called when the USB is ejected or removed from a PC
// Note this is from a USB IRQ so no printing to SerialUSB/etc.
void unplug(uint32_t i) {
  (void) i;
  okayToWrite = true;
}

// settings
struct configuration {
  byte testcheck;
  int ledbright;
  int ledspeed;
  int ledrotaterate;
  byte buttonB;
  byte buttonC;
  byte debugtoggle;
  byte valthres;
  byte baseval;
};
configuration configVar;

// ----------------------------------------------------------------------

void LEFT(){
  if((long)(micros() - last_micros) >= debouncing_time * 1000) {
    l_count ++;
    btncounter ++;
    btnid = 1;
    last_micros = micros();
  }
}
void RIGHT(){
  if((long)(micros() - last_micros) >= debouncing_time * 1000) {
    r_count ++;
    btncounter ++;
    btnid = 2;
    last_micros = micros();
  }
}

void cptch(){
  cap_sensor.update();
  
  if(cap_sensor.onChange()) {
    curval = cap_sensor.getValue();
    if (configVar.debugtoggle) { Serial.println(curval); }

    if (curval > (configVar.baseval + configVar.valthres)) { curpress++; }
    else { curpress = 0; }
    
    if (curpress > pressthres) {
      curpress=0;
      btncounter ++;
      btnid = 3;
      // future: setup timer for dont press again in x seconds
    }
  }
}

// ----------------------------------------------------------------------

void setup() {
  Keyboard.begin();
  Mouse.begin();
  Serial.begin(9600);

  // initialize the pushbutton pin as an input:
  // pinMode(buttonPin_1, INPUT);
  // pinMode(buttonPin_2, INPUT);
  // pinMode(LEFT_PIN, INPUT_PULLUP);
  pinMode(LEFT_PIN, INPUT);
  attachInterrupt(digitalPinToInterrupt(LEFT_PIN), LEFT, FALLING);
  pinMode(RIGHT_PIN, INPUT);
  attachInterrupt(digitalPinToInterrupt(RIGHT_PIN), RIGHT, FALLING);

  pixels.begin();
  ws2812fx.init();

  // start an initial loop to show it is on
  ws2812fx.setBrightness(20);
  ws2812fx.setSpeed(500);
  ws2812fx.setMode(FX_MODE_SCAN);
  ws2812fx.start();
  ledDelay.start(500);  // start a 1/2sec delay

  LittleFS.begin();
  singleFileDrive.onPlug(plug);
  singleFileDrive.onUnplug(unplug);
  singleFileDrive.begin("TheHackerMindset.URL", "TheHackerMindset.URL");

  Serial.println(F("Initialized."));
}

// ----------------------------------------------------------------------

void loadSettings(){
  EEPROM.begin(512);  //Set eeprom - 512 bytes
  EEPROM.get(0, configVar);
  if (configVar.testcheck != 0x04){
    //defaults
    configVar.testcheck = 0x04;      //psuedo check 'bit'
    configVar.ledbright = 7;                    
    configVar.ledspeed = 2000;                  
    configVar.ledrotaterate = 5000;
    configVar.buttonB = 2;             
    configVar.buttonC = 0;
    configVar.debugtoggle = 0;
    configVar.valthres = 10;
    configVar.baseval = 52;
    EEPROM.put(0, configVar);
    EEPROM.commit();
  }
  ws2812fx.setBrightness(configVar.ledbright);
  ws2812fx.setSpeed(configVar.ledspeed);
}

void saveSettings(){
  configVar.testcheck = 0x04; //make sure check 'bit' is set right
  configVar.ledbright = ws2812fx.getBrightness();
  configVar.ledspeed = ws2812fx.getSpeed();
  EEPROM.put(0, configVar); 
  EEPROM.commit();
}

bool runonce = true;
void checkInitLed() { // the led task
  // check if delay has timed out
  if (runonce) {
    ws2812fx.service();
  }
  if (ledDelay.justFinished()) { // only returns TRUE once.
    runonce = false;
    ws2812fx.stop();
    
    loadSettings();
    
    baselinerun = true;
    cptchbaselinetimer.start(2000); //start a two second timer
  }
}

// ----------------------------------------------------------------------
// serial control stuff

void recvChar() 
{
  static byte index = 0;
  while (Serial.available() > 0 && scmd_complete == false)
  {
    char rc = Serial.read();
    if (rc != '\n') {
      if(index < MAX_NUM_CHARS) scmd[index++] = rc;
    } else {
      scmd[index] = '\0'; // terminate the string
      index = 0;
      scmd_complete = true;
      if (configVar.debugtoggle) { Serial.print("received: '"); Serial.print(scmd); Serial.println("'"); }
    }
  }  
}

const char usageText[] PROGMEM = R"=====(
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
The Hacker Mindset Badge
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
LED Usage:
list modes    : list led modes
m <n>         : select mode <n>
r <n>         : set rotation rate on favorites to <n> ms (5000=5 secs)
s+            : increase speed
s-            : decrease speed
s <n>         : set speed to <n> (10=very fast, 5000=very slow)
b <n>         : set brightness to <n> (0=off, 40=full)
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Keyboard / Mouse Usage:
list payloads : list payloads
run <num>     : run payload number x
button2 <num> : set payload on button press
button3 <num> : set payload on button press
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
save          : save all current settings (leds, button mode)
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
)=====";

void printHW() {
  Serial.print(F("CPU: "));
  Serial.println(rp2040.f_cpu());
  Serial.print(F("Ram Used: "));
  Serial.println(rp2040.getUsedHeap());
  Serial.print(F("Ram Free: "));
  Serial.println(rp2040.getFreeHeap());
  Serial.print(F("Ram Total: "));
  Serial.println(rp2040.getTotalHeap());
}
void printUsage() {
  Serial.println((const __FlashStringHelper *)usageText);
}
void printModes() {
  Serial.println(F("LED pattern modes codes: "));
  for(int i=0; i < ws2812fx.getModeCount(); i++) {
    Serial.print(i);
    Serial.print(F("\t"));
    Serial.println(ws2812fx.getModeName(i));
  }
  Serial.println();
}
void printSettings() {
  Serial.print(F("LED current mode: "));
  Serial.print(ws2812fx.getMode());
  Serial.print(" : ");
  Serial.println(ws2812fx.getModeName(ws2812fx.getMode()));
  Serial.print(F("LED brightness: "));
  Serial.println(ws2812fx.getBrightness());
  Serial.print(F("LED speed: "));
  Serial.println(ws2812fx.getSpeed());
  Serial.print(F("LED rotation rate: "));
  Serial.println(configVar.ledrotaterate);
  Serial.print(F("Button 2 mode: "));
  Serial.println(configVar.buttonB);
  Serial.print(F("Button 3 mode: "));
  Serial.println(configVar.buttonC);
  Serial.print(F("Button 3 baseval: "));
  Serial.println(configVar.baseval);
  Serial.print(F("Button 3 threshold: "));
  Serial.println(configVar.valthres);
}
void printPayloads() {
  Serial.println(F("Keyboard / Mouse Payloads: "));
  Serial.println(F("0\tBrowser launch to website (windows)"));
  Serial.println(F("1\tMouse Jiggler"));
  Serial.println(F("2\tStatic Strings"));
  Serial.println(F("3\tKB Ducky - Windows"));
  Serial.println(F("4\tKB Ducky - Linux"));
}
void printDir() {
  Serial.println(F("FS:"));
  Dir dir = LittleFS.openDir("/");
  while (dir.next()) {
    Serial.println(dir.fileName());
  }
}
void catfile(String filepath) {
  Serial.println(filepath);
  File f = LittleFS.open(filepath, "r");
  while(f.available()) {
    String line = f.readStringUntil('\n');
    Serial.println(line);
  }
}

void runKBmode(byte t) {  
  if ((isKBloop) and (t == 1)){
      Serial.println("Halting mouse jiggler mode.");
  }
  else {
    Serial.print("Running Payload: ");
    Serial.println(t);  
  }
  
  if ((t == 0) or (t == 255)) { 
     RunPayload("DEFAULTDELAY 500");
     RunPayload("GUI r");
     RunPayload("STRING https://link.ws/hackermindsetdc31");
     RunPayload("ENTER");
     RunPayload("");
  }
  else if (t == 1){
    if (isKBloop)
    {
      isKBloop = 0; //turn off if triggered function again
    }
    else
    {
      mousejigglerDelay.start(5000);  // start a 5sec delay
      isKBloop = 1;
    }
  }
  else if (t == 2){
    // Keyboard.print("supersecretstring");
    RunPayload(F("STRING # this can be a secret key. visit https://link.ws/hackermindsetdc31 for more info #"));
  }
  else if (t == 3){
    //KB Ducky - windows
    RunPayloadFile("w-cmd-ipconfig.txt");
  }
  else if (t == 4){
    //KB Ducky - linux
    RunPayloadFile("l-xterm-id-ifconfig.txt");
  }  
}

void process_command() {
  int b;
  if (strncmp(scmd,"b ",2) == 0) {
    //uint8_t b = (uint8_t)atoi(scmd + 2);
    b = (int)atoi(scmd + 2);
    if (b > 40) { b = 40; }
    if (b < 1) { b = 1; }
    ws2812fx.setBrightness(b);
    Serial.print(F("Set brightness to: "));
    Serial.println(b);
  }
  else if (strcmp(scmd,"s+") == 0) {
    ws2812fx.setSpeed(ws2812fx.getSpeed() * 1.2);
    Serial.print(F("Increased speed by 20% to: "));
    Serial.println(ws2812fx.getSpeed());
  }
  else if (strcmp(scmd,"s-") == 0) {
    ws2812fx.setSpeed(ws2812fx.getSpeed() * 0.8);
    Serial.print(F("Decreased speed by 20% to: "));
    Serial.println(ws2812fx.getSpeed());
  }
  else if (strncmp(scmd,"s ",2) == 0) {
    uint16_t s = (uint16_t)atoi(scmd + 2);
    ws2812fx.setSpeed(s); 
    Serial.print(F("Set speed to: "));
    Serial.println(ws2812fx.getSpeed());
  }
  else if (strncmp(scmd,"c ",2) == 0) {
    uint32_t c = (uint32_t)strtoul(scmd + 2, NULL, 16);
    ws2812fx.setColor(c);
    Serial.print(F("Set color to: 0x"));
    Serial.println(ws2812fx.getColor(), HEX);
  }
  else if (strncmp(scmd,"r ",2) == 0) {
    int r = (int)atoi(scmd + 2);
    configVar.ledrotaterate = r;
    Serial.print(F("LED rotation rate: "));
    Serial.println(configVar.ledrotaterate);
  }
  else if (strcmp(scmd,"list modes") == 0) {
    printModes();
  }
  else if (strncmp(scmd,"m ",2) == 0) {
    uint8_t m = (uint8_t)atoi(scmd + 2);
    LED_WS2812FX(m);
    Serial.print(F("LED mode to: "));
    Serial.print(ws2812fx.getMode());
    Serial.print(" - ");
    Serial.println(ws2812fx.getModeName(ws2812fx.getMode()));
  }
  else if (strcmp(scmd,"off") == 0) {
    ws2812fx.stop();
    isLEDmode = 0;
  }
  
  else if (strcmp(scmd,"list payloads") == 0) {
    printPayloads();
  }
  else if (strncmp(scmd,"run ",4) == 0) {
    byte ru = (byte)atoi(scmd + 4);
    runKBmode(ru);
  }
  else if (strncmp(scmd,"button2 ",7) == 0) {
    byte ra = (byte)atoi(scmd + 8);
    configVar.buttonB = ra;
  }
  else if (strncmp(scmd,"button3 ",7) == 0) {
    byte rb = (byte)atoi(scmd + 8);
    configVar.buttonC = rb;
  }
  else if (strncmp(scmd,"thres ",5) == 0) {
    byte rc = (byte)atoi(scmd + 6);
    configVar.valthres = rc;
  }
  else if (strcmp(scmd,"baseline") == 0) {
    sampltotal = 0;
    samplcount = 0; 
    baselinerun = true;
    cptchbaselinetimer.start(2000); //start a two second timer
  }
  else if (strncmp(scmd,"base ",4) == 0) {
    byte rd = (byte)atoi(scmd + 5);
    configVar.baseval = rd;
  }
  else if (strcmp(scmd,"echo *") == 0) {
    printDir();
  }

  else if (strcmp(scmd,"?") == 0) {
    printUsage();
  }
  else if (strcmp(scmd,"cur") == 0) {
    printSettings();
  }
  else if (strcmp(scmd,"hwinfo") == 0) {
    printHW();
  }
  else if (strcmp(scmd,"save") == 0) {
    Serial.println(F("Saving settings:"));
    printSettings();
    saveSettings();
  }
  else if (strcmp(scmd,"reboot") == 0) {
    Serial.println(F("Rebooting..."));
    rp2040.reboot();
  }
  else if (strcmp(scmd,"d") == 0) {
    Serial.println("");
  }
  else if (strcmp(scmd,"debug") == 0) {
    if ((configVar.debugtoggle == 255) or (configVar.debugtoggle == 1)) { configVar.debugtoggle = 0; }
    else { configVar.debugtoggle = 1; }
  }

  scmd[0] = '\0';         // reset the commandstring
  scmd_complete = false;  // reset command complete
}

// ----------------------------------------------------------------------          
void loop() {
  checkInitLed();

  //if (!runonce) { cptch(); }
  if ((!runonce) && (!baselinerun)) { cptch(); }

  recvChar(); // read serial comm
  if(scmd_complete) { process_command(); }

  if (isKBloop) //for mouse jiggler mode
  {
    if (mousejigglerDelay.justFinished()) { 
      if (jigglerstage == 0) { Mouse.move(5,0); jigglerstage = 1;}
      else if (jigglerstage == 1) { Mouse.move(-5,0); jigglerstage = 2;}
      else if (jigglerstage == 2) { Mouse.move(0,5); jigglerstage = 3;}
      else if (jigglerstage == 3) { Mouse.move(0,-5); jigglerstage = 0;}
      mousejigglerDelay.start(5000); // start another one
    }
  }

  if (isLEDmode)
  {
    if (l_count == 1){
      //favorites / auto rotation
      if (ledrotateDelay.justFinished()) { 
        ws2812fx.setMode(favs[favs_current]);
        if (favs_current == favs_cnt) { favs_current = 0; }
        else { favs_current = favs_current + 1; }
        ledrotateDelay.start(configVar.ledrotaterate); // start another one
      }
    }
    ws2812fx.service();
  }

  if (baselinerun) {
    cap_sensor.update();
    if(cap_sensor.onChange()) {
        sampltotal = sampltotal + cap_sensor.getValue();
        samplcount++;
    }
    if (cptchbaselinetimer.justFinished()) { // only returns TRUE once.
      baselinerun = false;
      configVar.baseval = sampltotal / samplcount;
      Serial.print("new baseline: ");
      Serial.println(configVar.baseval);
    }
  }

  if (btncounter > 0)
  { 
    switch(btnid)
    {
      case 1:       
        // left button
        btncounter--;

        switch(l_count)
        {
          case 1:
            Serial.println(F("LED mode 1: Favorites auto rotate"));
            LED_WS2812FX(20);
            ledrotateDelay.start(configVar.ledrotaterate);  // start a 2.5sec delay
            break;
          case 2:
            Serial.println(F("LED mode 2: 20 Twinkle random"));
            LED_WS2812FX(20);
            break;
          case 3:
            Serial.println(F("LED mode 3: 10 Multi dynamic"));
            LED_WS2812FX(10);
            break;
          case 4:
            Serial.println(F("LED mode 4: 46 Fireworks random"));
            LED_WS2812FX(46);
            break;
          case 5:
            Serial.println(F("mode 5: Disable LED"));
            // reset to blank
            ws2812fx.stop();
            l_count = 0;
            isLEDmode = 0;
            break;
        }
        break;
      case 2:
        // right button
        btncounter--;
        Serial.println(F("Button 2 mode triggered:"));
        runKBmode(configVar.buttonB);
        break;
      case 3:
        // capacitive button
        btncounter--;
        Serial.println(F("Button 3 mode triggered:"));
        runKBmode(configVar.buttonC);
        break;
    }
  }
}
